import { Rules } from "./rules";
import Settings from "../settings";
import getLevelDB, { LevelDB, deleteLevelDB } from "../storage";
import DocumentLock from "./lock";
import { DocumentQuery, CollectionQuery, Query, QueryError } from "./query";
import Logging from "@hibas123/nodelogging";
import Session from "./session";
import nanoid = require("nanoid");

type IWriteQueries = "set" | "update" | "delete" | "add";
type ICollectionQueries = "get" | "add" | "keys" | "delete-collection" | "list";
type IDocumentQueries = "get" | "set" | "update" | "delete";

export interface ITypedQuery<T> {
   path: string[];
   type: T;
   data?: any;
   options?: any;
}

interface ITransaction {
   queries: ITypedQuery<IWriteQueries>[];
}

export type IQuery = ITypedQuery<ICollectionQueries | IDocumentQueries>;


export class DatabaseManager {
   static databases = new Map<string, Database>();

   static async init() {
      let databases = await Settings.getDatabases();

      databases.forEach(dbconfig => {
         let db = new Database(dbconfig.name, dbconfig.accesskey, dbconfig.rules, dbconfig.publickey, dbconfig.rootkey);
         this.databases.set(dbconfig.name, db);
      })
   }

   static async addDatabase(name: string) {
      if (this.databases.has(name))
         throw new Error("Database already exists!");

      await Settings.addDatabase(name);
      let database = new Database(name);
      this.databases.set(name, database);
      return database;
   }

   static getDatabase(name: string) {
      return this.databases.get(name);
   }

   static async deleteDatabase(name: string) {
      let db = this.databases.get(name)
      if (db) {
         await Settings.deleteDatabase(name);
         await db.stop();
         await deleteLevelDB(db.name)
      }
   }
}

export type ChangeTypes = "added" | "modified" | "deleted";

export type Change = {
   data: any;
   document: string;
   type: ChangeTypes;
   sender: string;
}


export class Database {
   private level = getLevelDB(this.name);

   get data() {
      return this.level.data;
   }

   get collections() {
      return this.level.collection;
   }


   public rules: Rules;
   public locks = new DocumentLock()
   public collectionLocks = new DocumentLock()

   public changes = new Map<string, Set<(change: Change) => void>>();

   toJSON() {
      return {
         name: this.name,
         accesskey: this.accesskey,
         publickey: this.publickey,
         rules: this.rules
      }
   }

   constructor(public name: string, public accesskey?: string, rawRules?: string, public publickey?: string, public rootkey?: string) {
      if (rawRules)
         this.rules = new Rules(rawRules);
   }

   async setRules(rawRules: string) {
      let rules = new Rules(rawRules);
      await Settings.setDatabaseRules(this.name, rawRules);
      this.rules = rules;
   }

   async setAccessKey(key: string) {
      await Settings.setDatabaseAccessKey(this.name, key);
      this.accesskey = key;
   }

   async setRootKey(key: string) {
      await Settings.setDatabaseRootKey(this.name, key);
      this.rootkey = key;
   }

   async setPublicKey(key: string) {
      await Settings.setDatabasePublicKey(this.name, key);
      this.publickey = key;
   }


   getQuery(path: string[], session: Session, type: "document" | "collection" | "any") {
      if (type === "document")
         return new DocumentQuery(this, path, session);
      else if (type === "collection")
         return new CollectionQuery(this, path, session);
      else
         return new Query(this, path, session);
   }

   private validate(query: ITypedQuery<any>) {
      const inv = new QueryError("Malformed query!");
      if (!query || typeof query !== "object")
         throw inv;

      if (!query.type)
         throw inv;

      if (!query.path)
         throw inv;
   }

   async run(query: IQuery, session: Session) {
      this.validate(query);
      const isCollection = query.path.length % 2 === 1;
      if (isCollection) {
         const q = new CollectionQuery(this, query.path, session);
         let type = query.type as ICollectionQueries;
         switch (type) {
            case "add":
               return q.add(query.data);
            case "get":
               const limit = (query.options || {}).limit;
               if (limit)
                  q.limit = limit;
               const where = (query.options || {}).where;
               if (where)
                  q.where = where;
               return q.get();
            case "keys":
               return q.keys();
            case "list":
               return q.collections();
            case "delete-collection":
               return q.deleteCollection();
            default:
               return Promise.reject(new Error("Invalid query!"));
         }
      } else {
         const q = new DocumentQuery(this, query.path, session);
         let type = query.type as IDocumentQueries;
         switch (type) {
            case "get":
               return q.get();
            case "set":
               return q.set(query.data, query.options || {});
            case "update":
               return q.update(query.data);
            case "delete":
               return q.delete();
            default:
               return Promise.reject(new Error("Invalid query!"));
         }
      }
   }

   async snapshot(query: ITypedQuery<"snapshot">, session: Session, onchange: (change: any) => void) {
      this.validate(query);

      const isCollection = query.path.length % 2 === 1;
      let q: DocumentQuery | CollectionQuery;
      if (isCollection) {
         q = new CollectionQuery(this, query.path, session);
         const limit = (query.options || {}).limit;
         if (limit)
            q.limit = limit;
         const where = (query.options || {}).where;
         if (where)
            q.where = where;
      } else {
         q = new DocumentQuery(this, query.path, session);
      }

      const id = nanoid(16);
      session.queries.set(id, q);
      return {
         id,
         snaphot: await q.snapshot(onchange)
      };
   }

   async unsubscribe(id: string, session: Session) {
      let query: CollectionQuery | DocumentQuery = session.queries.get(id) as any;
      if (query) {
         query.unsubscribe();
         session.queries.delete(id);
      }
   }

   async stop() {
      await this.data.close();
   }

   public async runCleanup() {
      const should = await new Promise<Set<string>>((yes, no) => {
         const stream = this.collections.iterator({
            keyAsBuffer: false,
            valueAsBuffer: false
         })

         const collections = new Set<string>();
         const onValue = (err: Error, key: string, value: string) => {
            if (err) {
               Logging.error(err);
               stream.end((err) => Logging.error(err))
               no(err);
            }

            if (!key && !value) {
               yes(collections);
            } else {
               collections.add(value)
               stream.next(onValue);
            }
         }

         stream.next(onValue);
      })

      const existing = await new Promise<Set<string>>((yes, no) => {
         const stream = this.data.iterator({
            keyAsBuffer: false,
            values: false
         })

         const collections = new Set<string>();
         const onValue = (err: Error, key: string, value: Buffer) => {
            if (err) {
               Logging.error(err);
               stream.end((err) => Logging.error(err))
               no(err);
            }

            if (!key && !value) {
               yes(collections);
            } else {
               let coll = key.split("/")[0];
               collections.add(coll)
               stream.next(onValue);
            }
         }

         stream.next(onValue);
      })

      const toDelete = new Set<string>();
      existing.forEach(collection => {
         if (!should.has(collection))
            toDelete.add(collection);
      })

      for (let collection of toDelete) {
         const batch = this.data.batch();

         let gt = Buffer.from(collection + "/ ");
         gt[gt.length - 1] = 0;

         let lt = Buffer.alloc(gt.length);
         lt.set(gt);
         lt[gt.length - 1] = 0xFF;

         await new Promise<void>((yes, no) => {
            const stream = this.data.iterator({
               keyAsBuffer: false,
               values: false,
               gt,
               lt
            })

            const onValue = (err: Error, key: string, value: Buffer) => {
               if (err) {
                  Logging.error(err);
                  stream.end((err) => Logging.error(err))
                  no(err);
               }

               if (!key && !value) {
                  yes();
               } else {
                  batch.del(key);
                  stream.next(onValue);
               }
            }

            stream.next(onValue);
         })

         await batch.write();
      }

      return Array.from(toDelete.values());
   }

}
